
c8y_bin_api = {}

function c8y_bin_api.c8y_url(path)
  local host = c8y.config.main:value("connection.server")
	host =  host:match("https?:\/\/.*\/")
	return path and (host .. path) or host
end


-----------------------------------------------------------------------------
-- Binary URL
--
-- Named (or not) args:
--  - id     - ID of uploaded file
--
function c8y_bin_api.url(id)
	id = id.id or id
	return c8y_bin_api.c8y_url('inventory/binaries/' .. id)
end


-----------------------------------------------------------------------------
-- Upload file to binary inventory
--
-- Named args:
--  - fileName         - file name or full path
--  - data             - file conent, if nil "fname" is read
--  - mime             - file mime type, default "application/octet-stream"
--  - boundary         - force boundary for multipart/form-data or nil
--  - timeout          - force timeout or nil
--  - onSuccess(id)    - function called on successful upload,
--                       id = binary file ID
--  - onError(e, msg)  - function called on upload failed,
--                       e   = error code,
--                       msg = error message
--  - onFinish()       - function called on upload finished
--                       (after onSuccess or onError)
--
function c8y_bin_api.upload(args)
	local fname = args.fileName or 'file'
	local boundary = args.boundary or string.format('SmartRestAgentBoundary7MA4YWx%06d', math.random(999999))
	local mime = args.mime or 'application/octet-stream'
	local data = args.data
	if not data then
		local file = io.open(fname)
		if not file then
			local err = 'Can not open file "' .. fname .. '"'
			print('WARNING: ' .. err)
			if args.onError then args.onError(-1, err) end
			return false, err
		end
		data = file:read('*a')
		file:close()
		fname = string.gsub(fname, "(.*/)(.*)", "%2")
	end
	local size = # data
	local conn = c8y.net:post(c8y_bin_api.c8y_url('inventory/binaries'))
	if args.timeout then conn.timeout = args.timeout end
	conn:setHeader('Accept: application/vnd.com.nsn.cumulocity.managedObject+json;')
	conn:setHeader('Content-Type: multipart/form-data; boundary=' .. boundary)
	conn:write('--' .. boundary .. '\r\n')
	conn:write('Content-Disposition: form-data; name="object"\r\n\r\n')
	conn:write('{"name":"' .. fname .. '","type":"' .. mime .. '"}\r\n')
	conn:write('--' .. boundary .. '\r\n')
	conn:write('Content-Disposition: form-data; name="filesize"\r\n\r\n')
	conn:write(size .. '\r\n')
	conn:write('--' .. boundary .. '\r\n')
	conn:write('Content-Disposition: form-data; name="file"; filename="' .. fname .. '"\r\n')
	conn:write('Content-Type: ' .. mime .. '\r\n\r\n')
	conn:write(data)
	conn:write('\r\n--' .. boundary .. '--\r\n')
	conn.onFinish = function ()
		local response = conn:read()
		if conn.ok then
			local id = response:match('"id"%s*:%s*"(%d+)"')
			if args.onSuccess then
				args.onSuccess(id)
			end
		end
		if args.onFinish then
			args.onFinish()
		end
	end
	conn.onError = function ()
		if args.onError then
			args.onError(conn.error, conn.errorMessage)
		end
	end
	conn:send()
	return true
end


-----------------------------------------------------------------------------
-- Download file from binary inventory
--
-- Named args:
--  - url       - file URL, or nil if "id" is used
--  - id        - ID file in binary inventory, used when no "url"
--  - fileName  - download file name  (full path), if nil temporary name is created
--  - timeout   - force timeout or nil
--  - onSuccess(fileName)  - function called on successful upload,
--                           fileName = downloaded fileName
--  - onError(e, msg)      - function called on upload failed,
--                           e   = error code,
--                           msg = error message
--  - onFinish()           - function called on upload finished
--                           (after onSuccess or onError)
--
-- Returns:
--  - nil, errMessage   - if failed, onERror is called too
--  - fileName          - on sucess returns used fileName
--
function c8y_bin_api.download(args)
	local url = args.url or args.id
	local fileName = args.fileName or string.format('/tmp/file%06d.tmp', math.random(999999))
	if not url then
		err = 'No download URL'
		print('WARNING: ' .. err)
		if args.onError then args.onError(-1, err) end
		return nil, err
	end
	if not url:find('/') then
		url = c8y_bin_api.url(url)
	end
	conn = c8y.net:get(url)
	conn:setOutFile(fileName)
	if args.timeout then conn.timeout = args.timeout end
	conn.onFinish = function ()
		if conn.ok then
			if file_exists(fileName) then
				if args.onSuccess then
					args.onSuccess(fileName)
				end
			else
				if args.onError then
					args.onError(-1, 'Creating local file "' .. fileName .. '" failed');
				end
			end
		end
		if args.onFinish then
			args.onFinish()
		end
	end
	conn.onError = function ()
		os.remove(fileName)
		if args.onError then
			args.onError(conn.error, conn.errorMessage)
		end
	end
	conn:send()
	return fileName
end

