
-- http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
-- http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b.pdf


--require "socket"
require "struct"
require "tools"

-- API
modbus = {}

-- some settings
CONNECT_TIMEOUT = 15
RECONNECT_INTERVAL = 5
REQUEST_TIMEOUT = 30


-- error codes
ERROR_TIMEOUT = 901


-----------------------------------------------------------------------------
-- ModBus addresses range
class "AddrRange"

-- Contruct range
function AddrRange:AddrRange(start, stop)
	self.start = start
	self.stop = stop or start
end

-- Get addresses count
function AddrRange:getCount()
	return self.stop - self.start + 1
end

-- Properties
function AddrRange:__index(key)
	if key == 'count' then return self:getCount() end
	return getmetatable(self)[key]
end


-----------------------------------------------------------------------------
-- ModBus message
class "ModBusMessage"

-- ModBus message constructor (internally used in ModBus class)
function ModBusMessage:ModBusMessage(args)
	self.id = args.id
	self.callback = args.callback
	self.data = ''
	self.unit = args.unit or 0
	self.len = args.len or 0
	self.ok = false
	self.error = nil
	self.sent = args.sent or ''
end

-- Unpack received data
function ModBusMessage:unpack(fmt)
	return self.ok and struct.unpack(fmt, self.data) or nil
end

-- Error message
function ModBusMessage.errorString(error)
	if type(error) ~= number and error ~= nil then error = error.error end
	return ModBus.errorString(error)
end


-----------------------------------------------------------------------------
-- Sender helper for ModBus devices
class "ModBusIo"

local ioConnections = {}

function ModBusIo:ModBusIo(args)
	self.server = args.server
	self.port = args.port
	self.id = args.id or (self.server .. tostring(self.port))
	self.sendQueue = {}
	self.messages = {}
	self.busy = false
	self.timeoutTimer = c8y.timer.new(REQUEST_TIMEOUT, function()
		self.busy = false
		local messages = self.messages
		self.messages = {}
		self:reconnect()
		for _,msg in pairs(messages) do
			msg.ok = false
			msg.error = ERROR_TIMEOUT
			if msg.timer then msg.timer:stop() msg.timer = nil end
			if msg.callback ~= nil then msg.callback(msg) end
		end
		-- self:sendFromQueue()   - synchronize lost, reconnect() used
	end)
	self.timeoutTimer.singleShot = true
end

-- send from Queue
function ModBusIo:sendFromQueue()
	-- DEBUG
	--function S(a,v) if v == nil then return tostring(a) end if a == nil then return '-' end return S(a[v]) end
	--local N = next(self.sendQueue)
	--DUMP('sendFromQueue(): busy='..S(self.busy)..', conn=(R='..S(self.conn,'running')..',C='..S(self.conn,'connected')..',F='..S(self.conn,'finished')..'), queue=(#'..#self.sendQueue..',N='..S(N)..')')

	if self.busy or next(self.sendQueue) == nil then return end
	if self.conn and self.conn.connected then
		self.busy = true
		self.timeoutTimer:start()
		local msg = table.remove(self.sendQueue, 1)
		DUMP('Sending msg '..msg.id..' to '..self.server..':'..self.port..':\n'..hex_dump(msg.sent))
		local conn = self:connection()
		if conn:send(msg.sent) then 
			self.messages[msg.id] = msg
		else
			-- reconnect
			table.insert(self.sendQueue, 1, msg)
			DUMP('send failed')
			self:reconnect()
		end
		--DUMP('Sent', id, msg)
	elseif not self.conn or not self.conn.running then
		DUMP('not running')
		self:reconnect()
	end
end

-- data to send
function ModBusIo:send(msg)
	table.insert(self.sendQueue, msg)
	msg.timer = c8y.timer.singleShot(REQUEST_TIMEOUT + CONNECT_TIMEOUT, function()
		DUMP('message '..msg.id..' timouted')
		for i,m in ipairs(self.sendQueue) do
			if m.id == msg.id then table.remove(self.sendQueue, i) break end
		end
		self.messages[msg.id] = nil
		msg.ok = false
		msg.error = ERROR_TIMEOUT
		if msg.timer then msg.timer:stop() msg.timer = nil end
		if msg.callback ~= nil then msg.callback(msg) end
		self:sendFromQueue()
	end)
	self:sendFromQueue()
end

-- receive
function ModBusIo:receive()
	--print('------ rrr -------')
	-- no connectiom nothing to receive
	if not self.conn then return end
	if self.conn.closed then self.conn = nil return end
	-- if we are inside of receiveing message, then receice data for this message
	if self.currentId then
		local msg = self.messages[self.currentId]
		assert(msg)
		r = self.conn:recv(msg.len - #msg.data)   -- header (reqId, proto, length, unit)
		--DUMP('RecvData', r)
		if r and #r > 0 then
			DUMP('Data recevied from '..self.server..':'..self.port..':\n'..hex_dump(r))
			-- collect data
			msg.data = msg.data .. r
			-- all data received?
			if #msg.data == msg.len then
				self.currentId = nil  -- finished
				self.timeoutTimer:stop()
				self.busy = false
				if #msg.data and msg.data:byte() < 0x80 then
					msg.ok = true
				else
					msg.ok = false
					msg.error = msg.data:byte(2)
				end
				if msg.timer then msg.timer:stop() msg.timer = nil end
				if msg.callback ~= nil then msg.callback(msg) end
				self.messages[msg.id] = nil
				self:sendFromQueue()
				return msg
			end
		end
	else
		-- if not inside message, then try receive header of new message
		r = self.conn:recv(7)   -- header (reqId, proto, length, unit)
		--DUMP('RecvHead', r)
		if r and #r > 0 then
			DUMP('Header received from '..self.server..':'..self.port..':\n'..hex_dump(r))
			local id, proto, len, unit = struct.unpack('!HHHB', r)
			local msg = self.messages[id]
			-- mesage must exists
			if msg == nil then error('Unknown response '..id)  self.conn:recv(len-1)  end
			--  len - 1, because unit is inside length section and we receive it already
			msg.len, msg.unit = len - 1, unit
			-- now we inside message receiving
			self.currentId = id
			-- try receive data now
			return self:receive()
		end
	end
end


-- return connection for operations
function ModBusIo:connection()
	if self.conn == nil then
		conn = c8y.net:tcp()
		conn.timeout = 5
		local host = self.server or c8y.config:value('server', '127.0.0.1')
		local port = self.port or c8y.config:value('port', 502)
		print('Connecting to '..host..':'..port)
		conn.timeout = CONNECT_TIMEOUT
		conn.timeout = self.__connectionTimeout
		self.conn = conn
		conn.onReceive = function()
			self:receive()
		end
		conn.onFinish = function()
			DUMP('ModBus ('..self.server..':'..self.port..') disconnected, error='..self.conn.error..' ('..self.conn.errorMessage..')')
			if self.conn.error then
				c8y.timer.singleShot(RECONNECT_INTERVAL, function() self:reconnect() end)
			else
				self:reconnect()
			end
		end
		conn.onConnect = function()
			DUMP('ModBus ('..self.server..':'..self.port..') connected')
			self:sendFromQueue()
		end
		local ok = conn:connect(host, port)
		if not ok then error('Connection to ModBus ('..self.server..':'..self.port..') failed: '..err) end
	end
	return self.conn
end

-- force reconnect
function ModBusIo:reconnect()
	if self.conn then
		self.conn.onFinish = nil
		self.conn:close()
	end
	self.busy = false
	self.timeoutTimer:stop()
	self.conn = nil
	return self:connection()
end

-- set connection timeout
function ModBusIo:setConnectionTimeout(timeout)
	self.__connectionTimeout = timeout
	if self.conn then self.conn.timeout = timeout end
end


-----------------------------------------------------------------------------
class "ModBus"

local ModBus_reqId = 1

-- Constructor, define ModBus object
-- args:
--   server  server to connect
--   port    port to conenct (TCP), default 502
function ModBus:ModBus(args)
	self.server = args.server or c8y.config:value('server', '127.0.0.1')
	self.port = args.port or c8y.config:value('port', 502)
	--self._reqId = 1   use ModBus_reqId
	self.proto = 0x0000                -- MODBUS protocol
	self.unit = args.unit or 0xff      -- Identification of a remote slave connected on a serial line or on other buses
	self.deviceId = args.deviceId
	self.deviceType = args.deviceType
	self.name = args.name or ''
	local ioid = self.server .. tostring(self.port) 
	if ioConnections[ioid] == nil then ioConnections[ioid] = ModBusIo{id=ioid, server=self.server, port=self.port} end
	self.io = ioConnections[ioid]
end

-- MedBus properies
function ModBus:__index(key)
	-- true if same messages precessing
	if key == 'waiting' then return next(self.messages) ~= nil end
	return getmetatable(self)[key]
end

-- Send message to cumulo
function ModBus:sendTemplate(args)
	return self.deviceType:send(args)
end

-- Add TCP ModBus layer (header)
function ModBus:addHeader(msg)
	local id = ModBus_reqId
	ModBus_reqId = ModBus_reqId + 1
	return  struct.pack('!HHHB', id, self.proto, #msg + 1, self.unit or 0xff) .. msg, id
end

-- set connection timeout
function ModBus:setConnectionTimeout(timeout)
	self.io:setConnectionTimeout(timeout)
end

-- send from Queue
function ModBus:_sendFromQueue()
	self.io:sendFromQueue()
end

-- send
function ModBus:send(msg, callback)
	local msg, id = self:addHeader(msg)
	msg = ModBusMessage{id=id, callback=callback, unit=self.unit, sent=msg, modbus=self}
	self.io:send(msg)
	return msg
end

-- receive
function ModBus:receive()
	return self.io:receive()
end

-- Error message
function ModBus.errorString(error)
	if error ==  0 or error == nil then return 'No error' end
	if error ==  1 then return 'Illegal function' end
	if error ==  2 then return 'Illegal data address' end
	if error ==  3 then return 'Illegal data value' end
	if error ==  4 then return 'Slave device failure' end
	if error ==  5 then return 'Acknowledge' end
	if error ==  6 then return 'Slave device busy' end
	if error ==  8 then return 'Memory parity error' end
	if error == 10 then return 'Gateway path unavailable' end
	if error == 11 then return 'Gateway target device failed to respond' end
	if error == ERROR_TIMEOUT then return 'Timeout' end
	return 'Unknown error ('..tostring(error)..')'
end

-- get addresses ranges
function ModBus:_addrRanges(addr, limit)
	table.sort(addr)
	local seq = {}   -- seq[]  - list of addr ranges (start, end)
	for _,a in ipairs(addr) do
		if #seq > 0 and seq[#seq].stop + 1 == a and seq[#seq].count < limit then seq[#seq].stop = a
		else seq[#seq+1] = AddrRange(a)
		end
	end
	return seq
end

-- Milti-get of coils / registers
function ModBus:_multiRead(args)
	local addr = args.addr
	local readFunc = args.readFunc
	local decodeFunc = args.decodeFunc
	local callback = args.callback
	local limit = args.limit or 64
	local values = args.values or {}
	local seq = {}
	if type(addr) == 'number' then
		local count = args.count or 1
		DUMP(count, type(count))
		for a = 1,count,limit do
			seq[#seq+1] = AddrRange(a+addr-1, math.min(a+addr+limit-2, addr+count-1))
		end
	else
		seq = self:_addrRanges(addr, limit)
	end
	local function cb(msg)
		if msg then
			if msg.error then
				if callback ~= nil then callback(nil) end
				return false
			end
			decodeFunc(values, msg)
		end
		if #seq > 0 then
			local range = table.remove(seq, 1)
			msg = readFunc(self, range.start, range.count, cb)
			msg.range = range;
			return true
		end
		if callback ~= nil then callback(values) end
	end
	return cb()
end

-- load coil
function ModBus:getCoil(args)
	return self:_multiRead{addr=args.addr, count=args.count, callback=args.callback, values=args.values, limit=2000,
		readFunc=function(self, addr, n, callback)
			return self:send(struct.pack('!BHH', 1, addr, n), callback)
		end,
		decodeFunc=function(values, msg)
			struct.unpack_bits_to_table{data=msg.data, from=3, count=msg.range.count, lsb=true, to=values}
		end
	}
end

-- save coil
function ModBus:setCoil(addr, val, callback)
	if type(val) == 'number' then val = val > 0 and true or false end
	return self:send(struct.pack('!BHH', 5, addr, (val and 0xff00 or 0)), callback)
end

-- (internal) load register worker
function ModBus:_getRegister(modbusMsg, args)
	return self:_multiRead{addr=args.addr, count=args.count, callback=args.callback, values=args.values, limit=125,
		readFunc=function(self, addr, n, callback)
			return self:send(struct.pack('!BHH', modbusMsg, addr, n), callback)
		end,
		decodeFunc=function(values, msg)
			for i=1,msg.range.count do
				values[#values+1] = struct.unpack_hint(msg.data, false, false, 1+i*2)
			end
		end
	}
end

-- load input register
function ModBus:getInputRegister(args)
	return self:_getRegister(4, args)
end

-- load input register
function ModBus:getHoldingRegister(args)
	return self:_getRegister(3, args)
end

-- save register
function ModBus:setRegister(addr, val, callback)
	return self:send(struct.pack('!BHH', 6, addr, val), callback)
end


-----------------------------------------------------------------------------
modbus.ModBus = ModBus
modbus.ModBusMessage = ModBusMessage
modbus.hex_dump = hex_dump

return modbus
