-- "struct" module
-- pack argument (half-int) to octet string  (big-endian)
--

-- API
struct = {}


-- Helper: pack half-int (16-bit) to octet string  (big-endian)
function struct.pack_hint(x, little)
	local b1 = x % 256  x = (x - b1) / 256
	local b2 = x % 256
	return little and string.char(b1, b2) or string.char(b2, b1)
end

-- Helper: pack int (32-bit) to octet string  (big-endian)
function struct.pack_int(x, little)
	local b1 = x % 256  x = (x - b1) / 256
	local b2 = x % 256  x = (x - b2) / 256
	local b3 = x % 256  x = (x - b3) / 256
	local b4 = x % 256
	return little and string.char(b1, b2, b3, b4) or string.char(b4, b3, b2, b1)
end

-- Helper: pack long-long-int (64-bit) to octet string  (big-endian)
-- TODO: check if Lua can hold long long integers (64bit)
function struct.pack_qint(x, little)
	local b1 = x % 0x100000000  x = (x - b1) / 0x100000000
	local b2 = x % 0x100000000
	if little then b1, b2 = b2, b1 end
	return pack_int(b2) .. pack_int(b1)
end

-- pack arguments to octet string like Python struct module
function struct.pack(fmt, ...)
	local little = string.byte(string.dump(function() end),7) > 0
	if #fmt == 0 then return '' end
	local out = ''
	local c = fmt:sub(1,1)
	local start = 2
	if c == '<' then little = true
	elseif c == '>' or c == '!' then little = false
	elseif c ~= '@' or c ~= '=' then start = 1
	end
	for i = start, #fmt do
		local c = fmt:sub(i,i)
		local v = select(i-start+1, ...)
		if      c == 'c' then out = out .. v:sub(1,1)
		elseif string.find('bB',   c, 1, true) then out = out .. string.char(v % 256)
		elseif string.find('hH',   c, 1, true) then out = out .. struct.pack_hint(v, little)
		elseif string.find('iIlL', c, 1, true) then out = out .. struct.pack_int(v, little)
		elseif string.find('qQ',   c, 1, true) then out = out .. struct.pack_qint(v, little)
		elseif string.find('sp',   c, 1, true) then out = out .. tostring(v)
		-- TODO: f, d, P
		end
	end
	return out
end


-- unpack argument (half-int) from octet string  (big-endian)
function struct.unpack_hint(s, little, sign, r)
	if r == nil then r = 1 end
	local b2, b1 = s:byte(r, r+1)
	if little then b1, b2 = b2, b1 end
	b1 = b2 * 256 + b1
	return (sign and b1 >= 0x8000) and (b1 - 0x10000) or b1
end

-- unpack argument (int) from octet string  (big-endian)
function struct.unpack_int(s, little, sign, r)
	if r == nil then r = 1 end
	local b4, b3, b2, b1 = s:byte(r, r+3)
	if little then b1, b2, b3, b4 = b4, b3, b2, b1 end
	b1 = b4 * 0x1000000 + b3 * 0x10000 + b2 * 256 + b1
	return (sign and b1 >= 0x80000000) and (b1 - 0x100000000) or b1
end

-- unpack arguments from octet string like Python struct module
function struct.unpack(fmt, s)
	local little = string.byte(string.dump(function() end),7) > 0
	if #fmt == 0 then return '' end
	local out = ''
	local c = fmt:sub(1,1)
	local start = 2
	if c == '<' then little = true
	elseif c == '>' or c == '!' then little = false
	elseif c ~= '@' or c ~= '=' then start = 1
	end
	local out, r = {}, 1
	local v
	for i = start, #fmt do
		if #s < r then return end
		local c = fmt:sub(i,i)
		if     c == 'x' then r = r + 1
		elseif c == 'c' then out[#out+1] = s:sub(r,r)  r = r + 1
		elseif c == 'B' then out[#out+1] = s:byte(r)  r = r + 1
		elseif c == 'b' then v = s:byte(r) out[#out+1] = (v > 127) and (v - 256) or v  r = r + 1
		elseif c == 'H' then out[#out+1] = struct.unpack_hint(s, little, false, r)  r = r + 2
		elseif c == 'h' then out[#out+1] = struct.unpack_hint(s, little, true, r)  r = r + 2
		elseif c == 'I' then out[#out+1] = struct.unpack_int(s, little, false, r)  r = r + 4
		elseif c == 'i' then out[#out+1] = struct.unpack_int(s, little, true, r)  r = r + 4
		--elseif string.find('qQ',   c, 1, true) then out = out .. pack_qint(v, little)
		elseif string.find('sp',   c, 1, true) then out[#out+1] = s:sub(r)  break
		-- TODO: f, d, P   and   q, Q
		end
	end
	return unpack(out)
end

-- unpack bits from octet string
function struct.unpack_bits_to_table(args)
	local s = args.data
	local from = args.from or 1
	local n = args.count or (#s * 8)
	local out = args.to or {}
	local lsb = args.lsb or false
	local i = 0
	while i < n do
		local b = 0
		local c = s:byte(i/8+from)
		while b < 8 and b + i < n do
			if lsb then
				v = (c % 2) > 0
				c = c / 2
				c = c - (c % 1)
			else
				v = c >= 128
				c = (c * 2) % 256
			end
			b = b + 1
			out[#out + 1] = v
		end
		i = i + 8
	end
	return out
end

-- unpack bits from octet string
function struct.unpack_bits(s, n, lsb)
	return unpack(struct.unpack_bits_to_table{data=s, count=n, lsb=lsb})
end



return struct
