
-- from http://lua-users.org/wiki/StringTrim
function string:strip()
	return self:match'^%s*(.*%S)' or ''
end


-- simple, from http://lua-users.org/wiki/SplitJoin
function string:split_simple(sep, strip)
	local sep, fields = sep or " \t\r\n\v", {}
	local pattern = string.format("([^%s]+)", sep)
	self:gsub(pattern, function(c) fields[#fields+1] = strip and c:strip() or c end)
	return fields
end


-- from http://lua-users.org/wiki/SplitJoin
function string:split(delim, maxNb, strip)
	-- Python default separator
	if delim == nil then
		delim = '[ \t\r\n\v]'
		str = self:strip()
	else
		str = self
	end
	-- Eliminate bad cases...
	if string.find(str, delim) == nil then
		return { strip and str:strip() or str }
	end
	if maxNb == nil or maxNb < 1 then
		maxNb = 0    -- No limit
	end
	local result = {}
	local pat = "(.-)" .. delim .. "()"
	local nb = 0
	local lastPos
	for part, pos in string.gfind(str, pat) do
		nb = nb + 1
		result[nb] = strip and part:strip() or part
		lastPos = pos
		if nb == maxNb then break end
	end
	-- Handle the last field
	if nb ~= maxNb then
		result[nb + 1] = string.sub(str, lastPos)
		if strip then result[nb + 1] = result[nb + 1]:strip() end
	end
	return result
end


-- from http://lua-users.org/wiki/StringRecipes
function string:startswith(prefix)
	return string.sub(self, 1, string.len(prefix)) == prefix
end


-- from http://lua-users.org/wiki/StringRecipes
function string:endswith(suffix)
	return suffix=='' or string.sub(self, -string.len(suffix)) == suffix
end


-- from http://stackoverflow.com/a/1413919
-- return a new array containing the concatenation of all of its 
-- parameters. Scaler parameters are included in place, and array 
-- parameters have their values shallow-copied to the final array.
-- Note that userdata and function values are treated as scalar.
function concat_to_table(...) 
    local t = {}
    for n = 1,select("#",...) do
        local arg = select(n,...)
        if type(arg)=="table" then
            for _,v in ipairs(arg) do
                t[#t+1] = v
            end
        else
            t[#t+1] = arg
        end
    end
    return t
end

-- from http://stackoverflow.com/questions/4990990/lua-check-if-a-file-exists
-- Return true if file exists a can be opened
function file_exists(name)
   local f=io.open(name,"r")
   if f~=nil then io.close(f) return true else return false end
end

-- some tricks from http://stackoverflow.com/questions/1340230/check-if-directory-exists-in-lua
-- Return true if entry (file or dir) exists
function file_entry_exists(name)
	if type(name)~="string" then return false end
	return os.rename(name,name) and true or false
end


-- author: Robert Kalinowski
-- Apply default funstion paramters (arguments)
-- Arguments:
--   params - function paramters (table)
--   def    - named and unnamed default paramters (table)
--            e.g.  def={x=3, y=7, z=0, 'first', 'second'}
--   order  - order of named parameters (table, optional)
--
-- function foo(params)
--   param = default_params{params, def={radius=5, 0, 0, 0}, order={'x','y','z'}}
--   ...
-- end
--
-- foo(1, 2, 3)     -> x=1, y=2, z=3, radius=5, [1]=1, [2]=2, [3]=3 
-- foo(1, 2, 3, 4)  -> x=1, y=2, z=3, radius=4, [1]=1, [2]=2, [3]=3 
-- foo(1, 2)        -> x=1, y=2, z=0, radius=4, [1]=1, [2]=2, [3]=0 
--
-- If you define named and unamed default value at the same time the named is used,
-- e.g.  {params, def={z=1}, 2}, order={'z'}}
--       bar(3) -> 3, [1] = 3,      bar() -> 1, [1] = 1
--
function default_params(args)
  -- helper of default_params() default arguments
  local ai = 1
  local function next_arg()
    local ret = args[ai]
    ai = ai + 1
    return ret
  end
  -- arguments of default_params()
  local arg = args.params or args.args or next_arg() or {}
  local def = args.def or next_arg() or {}
  local ord = args.order or next_arg() or {}
  -- unnamed argument as named in order from 'names'
  for i = 1,#ord do
    local n = ord[i]
    if arg[i] == nil then arg[i] = (arg[n] == nil) and def[n] or arg[n] end
    if arg[n] == nil then arg[n] = (arg[i] == nil) and def[i] or arg[i] end
  end
  -- apply defaut arguments
  for k,v in pairs(def) do
    if arg[k] == nil then arg[k] = v end
  end
  -- return filled arguments
  return arg
end



-- from http://www.lua.org/pil/19.3.html
-- Iterator in key order
function pairsByKeys (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end


-- Table keys
function table.keys(tab, sort)
	local out = {}
	for k,_ in pairs(tab) do table.insert(k) end
	if sort then table.sort(out) end
	return out
end

-- Nice hex dump to IO
function hex_dump_io(buf)
	for byte=1, #buf, 16 do
		local chunk = buf:sub(byte, byte+15)
		io.write(string.format('%08X  ',byte-1))
		chunk:gsub('.', function (c) io.write(string.format('%02X ',string.byte(c))) end)
		io.write(string.rep(' ',3*(16-#chunk)))
		io.write(' ',chunk:gsub('%c','.'),"\n") 
	end
end

-- Return nice hex dump
function hex_dump(buf)
	out = ''
  for byte=1, #buf, 16 do
    local chunk = buf:sub(byte, byte+15)
    out = out .. string.format('%08X  ',byte-1)
    chunk:gsub('.', function (c) out = out .. string.format('%02X ',string.byte(c)) end)
    out = out .. string.rep(' ',3*(16-#chunk))
    out = out .. ' ' .. chunk:gsub('%c','.') .. "\n"
  end
	return out
end

