require "tools"
require "c8y_bin_api"

local opFile = c8y.config:value("opFile", "/var/lib/sragent/configFile.op")
local inProgress = false
local tempFile = c8y.config:value("tempFile", "/opt/cdcs/upload/configFile.cfg.tar.gz")
local templates = {}

local debug = c8y.config:boolValue("debug") and print or function() end


-----------------------------------------------------------------------
-- Save operation ID to finish it after reboot
local function saveOperation(id)
  local file = io.open(opFile, "w+")
  file:write(id)
  file:close()
end

-----------------------------------------------------------------------
-- Install configuration
local function installFile(fileName)
	fileName = fileName or tempFile
  debug('INSTALL FILE "' .. fileName .. '"')
  local file = io.popen("install_file " .. fileName)
  local response = file:read("*a")
	debug('INSTALING...', response, '###')
  return response:find("is successful") ~= nil
end

-----------------------------------------------------------------------
-- Restart the device
local function reboot()
  debug("REBOOT")
  os.execute("reboot")
end

-----------------------------------------------------------------------
-- Handle download operation
function download(op)
  if inProgress then
    op:setFailed("Config operation already in progress")
  else
    inProgress = true
    local url = op:getResponseRecord(1)[3]
    local configDumpId = op:getResponseRecord(1)[4]
    debug("EXECUTING DOWNLOAD HANDLER", op.id, url, configDumpId)
		c8y_bin_api.download{
			url = url,
			fileName = tempFile,
			onSuccess = function(fileName)
				saveOperation(op.id, configDumpId)
				if installFile(fileName) then
					templates.updateConfigDumpFragment:send{c8y.device.id, configDumpId}
					reboot()
				else
					debug("Installing the config file failed")
					op:setFailed("Installing the config file failed")
					os.remove(opFile)
					inProgress = false
				end
			end,
			onError = function(e, errorText)
				debug("Download failed: " .. (errorText or "???"))
				op:setFailed(errorText or "Unknown error")
				os.remove(opFile)
				inProgress = false
			end
		}
  end
end

-----------------------------------------------------------------------
-- Export config to the file
local function exportConfig()
  local fileName = string.format("NTC-6200_Settings_%d_%s.cfg", c8y.device.id, c8y.strftime('%Y%m%d'))
	local filePath = '/opt/cdcs/upload/' .. fileName .. '.tar.gz'
  if os.execute("dbcfg_export -o /tmp/" .. fileName ..  ' -p ""') ~= 0 then
    return nil, "Exporting rdb config failed"
  end
	local openvpnkeys = file_entry_exists('/usr/local/cdcs/openvpn-keys') and 'openvpn-keys ' or ''
	local sshhostkeys = file_entry_exists('/usr/local/cdcs/ssh-hostkeys') and 'ssh-hostkeys ' or ''
  if  os.execute('tar -C /usr/local/cdcs -zcvf - ipsec.d ' .. openvpnkeys .. sshhostkeys .. ' | openssl des3 -salt -k "" | dd of=/tmp/vpn.des3') ~= 0 then
    return nil, "Exporting vpn config failed"
  end
  if os.execute("cd /tmp/ && tar -zcvf " .. filePath .. " " .. fileName .. " vpn.des3") ~= 0 then
    return nil, "Archiving config failed"
  end
	os.remove('/tmp/' .. fileName)
  return filePath
end

-----------------------------------------------------------------------
-- Create configuration dump MO
local function createConfigDump(name, description, binaryId)
  local url = c8y_bin_api.url(binaryId)
  local response = templates.createConfigRepoEntry:send{name, description, url}
	local configDumpId = response:getRecord(0, templates.response.id)[3]
	if configDumpId then
		return configDumpId
	end
  return nil, "Error creating configuration repository entry: response invalid"
end

-----------------------------------------------------------------------
-- Update current configuration fragment in the device
local function updateConfigDumpFragment(configDumpId)
  local response = templates.updateConfigDumpFragment:send({c8y.device, configDumpId})
  if response.httpCode == 200 then
    return true
	end
	return false, "Error updating configuration dump device fragment"
end

-----------------------------------------------------------------------
-- Handle upload operation
function upload(op)
  debug("EXECUTING UPLOAD HANDLER", op.id)
  local fileName, errorText = exportConfig()
  if not fileName then
    op:setStatus(op.Failed, errorText)
    return
  end

	c8y_bin_api.upload{
		fileName = fileName,
		onSuccess = function(binaryId)
			local url = c8y_bin_api.url(binaryId)
			local name = "NTC 6200 Device Configuration"
			local descr = string.format("Uploaded by device %d: %s at %s",
				c8y.device.id, c8y.device.defaultName, c8y.strftime('%F %T'))
			local response = templates.createConfigRepoEntry:send{name, descr, url}
			if response.ok then
				local configDumpId = response:getRecord(0, templates.response.id)[3]
				if configDumpId then
					local success, errorText = updateConfigDumpFragment(configDumpId)
					op:setStatus(success and op.Successful or op.Failed, errorText)
					return
				end
			end
			op:setFailed('Creating config repo entry failed')
		end,
		onError = function(err, msg)
			op:setFailed('Upload fieled: ' .. msg)
		end}
end

-----------------------------------------------------------------------
-- Plugin initialize
function init()
  debug("CONFIG FILE PLUGIN DEBUGGING ENABLED")
	-- Templates
  c8y.operations:addOperation("c8y_DownloadConfigFile","$.url,$.c8y_ConfigurationDump.id", download)
  c8y.operations:addOperation{'c8y_UploadConfigFile', '', upload}
  templates.createConfigRepoEntry = c8y.templates:addRequest('POST,/inventory/managedObjects,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,STRING STRING STRING,"{""name"":""%%"",""description"":""%%"",""url"":""%%"",""type"":""c8y_ConfigurationDump""}"')
  templates.updateConfigDumpFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,UNSIGNED UNSIGNED,"{""c8y_ConfigurationDump"":{""id"":""%%""}}"')
  templates.response = c8y.templates:addResponse(',$.id,$.id')
end

-----------------------------------------------------------------------
-- Plugin is atarting
function start()
	-- check if maybe we are executed after reboot during install operation
  local file = io.open(opFile)
  if file then
    debug("OP FILE EXISTS")
    local id = file:read("*n")
    file:close()
    debug("ID FROM OP FILE", id)
    if id then
      c8y.operations:setOperationStatus(id, c8y.operations.Successful)
    end
    os.remove(opFile)
  end
end




-----------------------------------------------------------------------
--- XXX   Tests   XXX  ---
if c8y.config:value("xxxTests") then
exportConfig = function()
	DUMP('Make configuration dump...')
  local fileName = string.format("NTC-6200_Settings_%d_%s.cfg.tar.gz", c8y.device.id, c8y.strftime('%Y%m%d'))
	os.execute('echo -n 12345 > ' .. fileName)
  return fileName 
end
installFile = function(fileName)
	DUMP('\nINSTALLING "' .. (filename or '') .. '"...\n')
	return true
end
reboot = function()
  DUMP('\n\n        REBOOT !!!   \n\n')
end
end   ---  Tests

