
require "inifile"
require "c8y_bin_api"

local inifname = c8y.config:value('conf')
local ini
local startTime = os.time() - io.open('/proc/uptime', 'r'):read('*n')
local templates = {}
local logs = {}


-- Filter msg
-- Available rules for one message:
--  - "dateFrom"  - UNIX time (number)
--  - "dateTo"    - UNIX time (number)
--  - "text"      - substring to find
function filter_log_message(msg, filter)
	if msg.unixtime then
		if filter.dateFrom and msg.unixtime < filter.dateFrom then return end
		if filter.dateTo   and msg.unixtime > filter.dateTo   then return end
	end
	if filter.line and not (msg.line and msg.line:find(filter.line)) then return end
	if filter.text and not (msg.text and msg.text:find(filter.text)) then return end
	return msg
end


-- Parse one log
function parse_log_line(sect, fields, line)
	local msg = {}
	-- parse line (using pattern)
	local v = {line:match(sect.pattern)}
	if #v == 0 then return end
	-- set all fileds (convert subexpression to named table)
	for i,f in ipairs(fields) do
		msg[f] = msg[f] and (msg[f]..' '..v[i]) or v[i]
	end
	-- parse / update date-time
	local t
	if sect.timeFormat == '$(uptime)' then
		t = tonumber(msg.time) + startTime
	else
		t = c8y.strptime(sect.timeFormat or '%F %T', msg.time, sect.timezone)
	end
	msg.unixtime = t
	if t then msg.time = c8y.strftime(ini.timeFormat, t) end
	msg.line = line
	return msg
end


-- Parse one log
-- Available filters rules:
--  - all from filter_log_message()
--  - "start"  - first message index (default 1), from the end if negative
--  - "count"  - number of messages (default 0, means all)
-- Mode:
--  - "string"   - whole entries in one string
--  - "lines"    - list of lines
--  - nil        - list of entries (with seprate fileds)
function parse_log(args)
	local name = args.name
	local filter = args.filter
	local mode = args.mode
	if not name then
		print('WARNING: no log name')
		return
	end
	local modeText = (mode == 'string' or mode == 'lines')
	local log = {}
	local lastMsg
	sect = ini[name]
	if sect then
		-- read data from log source
		local src
		if sect.file then 
			src = io.lines(sect.file)
		elseif sect.execute then
			src = io.popen(sect.execute):lines()
		end
		-- parse log (if exists)
		if src then
			local fields = sect.fields:split(',', 0, true)
			for line in src do
				-- parse each line
				local msg = parse_log_line(sect, fields, line)
				-- line continuation?
				if not msg and lastMsg then
					if modeText then
						log[#log] = log[#log] .. '\n' .. line
					else
						lastMsg.line = lastMsg.line .. '\n' .. line
					end
				else
					-- filter line (if filter exists)
					if msg and filter then
						msg = filter_log_message(msg, filter)
						lastMsg = msg
					end
					-- add message to log (if not filtered)
					if msg then
						log[#log + 1] = modeText and msg.line or msg
						lastMsg = msg
					end
				end
			end
		else
			print('WARNING: no source in log "'..name..'"')
			return
		end
	else
		print('WARNING: no log section "'..name..'"')
		return
	end
	-- filter log entries (limit range)
	if filter then
		local start = filter.start or 1
		local count = filter.count or 0
		local N = #log
		if start == 0 then start = 1 elseif start < 0 then start = N + start + 1 end
		local stop = (count < 1) and N or (start + count - 1)
		if start < 1 then start = 1 end
		if stop > N then stop = N end
		--DUMP('#', start, stop, N)
		if start > 1 or stop < N then
			local old = log
			log = {}
			for i = start, stop do
				log[#log+1] = old[i]
			end
		end
	end
	if mode == 'string' then
		log = table.concat(log, '\n')
	end
	return log
end


-- Initialize plugin
function init()
	-- load ini file
	if not inifname then
		error('No "conf" value in configuration')
	end
	if not file_exists(inifname) then
		error('Configuration file "'..inifname..'" does not exist')
	end
	ini = inifile.parse(inifname)
	-- suppoprted logs
	print('Supported logs: ' .. ini.logs)
	for _,name in ipairs(ini.logs:split(',', 0, true)) do
		if ini[name] and ini[name].pattern and ini[name].fields then
			logs[#logs + 1] = name
		end
	end
	-- templates
	templates.id = c8y.templates:new()
	templates.supportedLogs = templates.id:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,,"{""c8y_SupportedLogs"":[%%]}"')
	-- operation
	c8y.operations:addOperation{'c8y_LogfileRequest', '$.logFile,$.dateFrom,$.dateTo,$.maximumLines,$.searchText', get_log, successfulFragment='"c8y_LogfileRequest":{"logFile":"%STRING%","dateFrom":"%STRING%","dateTo":"%STRING%","maximumLines":%INTEGER%,"searchText":"%STRING%","file":"%STRING%"}'}

	-- XXX tests...
	if false then
		for _,name in ipairs(logs) do
			filter = {}
			--filter = {start=-3, count=2}
			--filter = {start=-1000}
			filter = {count=15}
			--filter.dateFrom = c8y.strptime('%FT%T%z', '2015-09-02T14:45:55+0200')
			--filter.dateTo   = c8y.strptime('%FT%T%z', '2015-09-02T14:45:56+0200')
			--filter.text = 'ACPI'
			filter.line = ' Agent'
			DUMP('filter:', filter)
			local log = parse_log{name=name, filter=filter, mode='string'}
			DUMP(#log, log)
		end
		error "XXXXXXXXXXXXX"
	end
end


-- Handle c8y_LogfileRequest operation
function get_log(op)
	local rec = op.responseRecord
	DUMP('Log request', rec)
	-- get filter values
	local logFile, filter = rec[3], {}
	local dateFrom, dateTo
	dateFrom, dateTo, filter.start, filter.line = rec[4], rec[5], - rec[6], rec[7]
	if dateFrom and #dateFrom then
		filter.dateFrom = c8y.strptime('%FT%T%z', dateFrom)
		if not filter.dateFrom then dateFrom = nil end
	end
	if dateTo and #dateTo then
		filter.dateTo = c8y.strptime('%FT%T%z', dateTo)
		if not filter.dateTo then dateTo = nil end
	end
	-- find log
	local error = 'Log "' .. logFile .. '" not supported'
	for _,log in ipairs(logs) do
		if log == logFile then
			-- parse log
			log = parse_log{name=logFile, filter=filter, mode='string'}
			if log then
				-- upload log
				error = nil
				print('Log data size: ' .. #log)
				c8y_bin_api.upload{data=log, fileName=logFile..'.log', mime='text/plain', onSuccess=
					function(id)
						-- upload success, finish the operation
						local url = c8y_bin_api.c8y_url("inventory/binaries/" .. id)
						op.successful = { logFile, dateFrom or '', dateTo or '', - filter.start, filter.line or '', url }
					end, onError=
					function(err, msg)
						-- upload failed, fail the operation
						op:setFailed('Upload log failed: ' .. msg)
					end}
			else
				error = 'Log "' .. logFile .. '" parse failed'
			end
		end
	end
	if error then
		op:setFailed(error)
	end
end


function start()
	-- present evailable logs
	local list = ''
	for _,log in ipairs(logs) do list = string.format('%s%s"%s"', list or '', #list>0 and ',' or '', log) end
	--DUMP('Supported logs:', list)
	templates.supportedLogs:send{c8y.device, list, JsonEscape=false}
	--DUMP('++++++++++++++++++++++++++++++++++++++++++++++++++')
end

