require("tools")

local restartEnable = c8y.config:value("restart.enable", "1")
local restartCommand = c8y.config:value("restart.command", "reboot")
local resetEnable = c8y.config:value("factoryReset.enable", "1")
local resetCommand = c8y.config:value("factoryReset.command", "rdb_set service.system.reset 1")
local rebootOpFile = c8y.config:value("restart.file", "/var/lib/sragent/reboot.op")

local templates = {}

local debugEnable = c8y.config:boolValue("debug") or false
local debug

if debugEnable then
  debug = print
else
  debug = function() end
end

local function isEnabled(configValue)
  debug("IS ENABLED", configValue, type(configValue))
  if configValue == "1" or configValue == 1 then
    return true
  else
    return false
  end
end

local function sendSavedOperations()
  debug("SEND SAVED OPERATIONS")
  local file = io.open(rebootOpFile)
  if file then
    local ids = {}
    local id = file:read("*n")
    while id do
      c8y.operations:setOperationStatus(id, c8y.operations.Successful)
      id = file:read("*n")
    end
    file:close()
    os.remove(rebootOpFile)
  end
end

local function sendRestartFragment()
  debug("SEND RESTART FRAGMENT" , restartEnable)
  if isEnabled(restartEnable) then
    templates.putRestartFragment:send({c8y.device})
  else
    templates.removeRestartFragment:send({c8y.device})
  end
end

local function sendResetFragment()
  debug("SEND RESET FRAGMENT" , resetEnable)
  if isEnabled(resetEnable) then
    templates.putResetFragment:send({c8y.device})
  else
    templates.removeResetFragment:send({c8y.device})
  end
end

local function watchEnableCofig()
  local oldRestartEnable = restartEnable
  local oldResetEnable = resetEnable
  c8y.timer.start(10, function()
    debug("CONFIG WATCHER")
    restartEnable = c8y.config:value("restart.enable", "1")
    resetEnable = c8y.config:value("reset.enable", "1")
    if restartEnable ~= oldRestartEnable then
      sendRestartFragment()
      oldRestartEnable = restartEnable
    end
    if resetEnable ~= oldResetEnable then
      sendResetFragment()
      oldResetEnable = resetEnable
    end
  end)
end

--------------------------------------------------------------------
class("RestartOperationHandler")

function RestartOperationHandler:operationExecuting(op)
  debug("EXECUTING", "RESTART", op.id)
  if isEnabled(restartEnable) then
    local file  = io.open(rebootOpFile,"a+")
    local previousText = file:read("*a")
    file:write("\n" .. op.id)
    file:close()
    local errorCode = os.execute(restartCommand)
    if errorCode ~= 0 then
      op:setStatus(op.Failed, 'Executing restart command "' .. (restartCommand or "") .. '" failed, error code: ' .. errorCode)
      if previousText and #previousText > 0 then
        file = io.open(rebootOpFile,"w+")
        file:write(previousText)
        file:close()
      else
        os.remove(rebootOpFile)
      end
    end
  else
    op:setStatus(op.Failed, "Restarting not enabled on device side")
  end
end
--------------------------------------------------------------------
class("ResetOperationHandler")

function ResetOperationHandler:operationExecuting(op)
  debug("EXECUTING", "RESET", op.id)
  if isEnabled(resetEnable) then
    local errorCode = os.execute(resetCommand)
    if errorCode ~= 0 then
      op:setStatus(op.Failed, 'Executing restart command "' .. (resetCommand or "") .. '" failed, error code: ' .. errorCode)
    else
      op:setStatus(op.Successful)
    end
  else
    op:setStatus(op.Failed, "Resetting not enabled on device side")
  end
end
--------------------------------------------------------------------

function init()
  templates.putRestartFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,,"{""c8y_Restart"":{}}"')
  templates.removeRestartFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,,"{""c8y_Restart"":null}"')
  templates.putResetFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,,"{""c8y_FactoryReset"":{}}"')
  templates.removeResetFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,,"{""c8y_FactoryReset"":null}"')
  local restartHandler = RestartOperationHandler()
  c8y.operations:addOperation("c8y_Restart", "", restartHandler)
  local resetHandler = ResetOperationHandler()
  c8y.operations:addOperation("c8y_FactoryReset", "", resetHandler)
end

function start()
  sendResetFragment()
  sendRestartFragment()
  watchEnableCofig()
  sendSavedOperations()
end
