require('tools')

-- name for package commands:
ipkg = 'ipkg-cl'
-- current arch:
arch = 'arm'
-- repo url:
repo = c8y.config:value('repo')
-- path to packages local storage
storage = c8y.config:value('storage', '/opt/cdcs/upload')
-- software operation backup file
swOpBackup = c8y.config:value('swOpBackup', "/opt/software_operation.txt")
-----------------------------------------------------------------------------
-- compare two versions; return like strcmp() in C:
--  < 0   -  v1 < v2
--  = 0   -  v1 == v2
--  > 0   -  v1 > v2
function compare_ver(v1, v2)
  v1 = tostring(v1)
  v2 = tostring(v2)
  if not v1 then
    v1 = ""
  end
  if not v2 then
    v2 = ""
  end
  local pattern = "([^.]+)"
  local splitV1 = v1:gmatch(pattern)
  local splitV2 = v2:gmatch(pattern)
  local done = false
  while not done do
    local matchV1 = splitV1()
    local matchV2 = splitV2()
    matchV1 = tonumber(matchV1)
    matchV2 = tonumber(matchV2)
    if not matchV1 and not matchV2 then
      done = true
    end
    if not matchV1 then
      matchV1 = 0
    end
    if not matchV2 then
      matchV2 = 0
    end
    if matchV1 ~= matchV2 then
      return matchV1 - matchV2
    end
  end
  return 0
end

-- Run the package command and fetch the command output
-- Global variable `ipkg' is used as command name
function command(arg)
  local cmd = ipkg or 'ipkg-cl'
  local h = io.popen(cmd..' '..arg)
  local res = h:read("*a")
  h:close()
  local res, cnt = res:gsub('^Successfully terminated', '')
  if cnt > 0 then return '' end
  res = res:gsub('\nSuccessfully terminated.*$', '')
  return res
end

function prepareSavedOperation(id, status, failureReason)
  savedOperationId = id
  savedOperationStatus = status
  savedOperationFailureReason = failureReason
end

function sendSavedOperationStatus()
  if savedOperationId and savedOperationStatus then
    if savedOperationFailureReason then
      c8y.operations:setOperationStatus(savedOperationId, savedOperationStatus, savedOperationFailureReason)
    else
      c8y.operations:setOperationStatus(savedOperationId, savedOperationStatus)
    end
  end
end

-----------------------------------------------------------------------------
-- Package class
class "Package"

-- Package constructor
--  name - the packages name
--  ver  - the package version
function Package:Package(name, ver, url, type) 
  self.name = name
  self.ver = ver
  self._url = url
  self.type = type or "ipk"
end

-- Properties
function Package:__index(key)
  if key == 'filename' then return self:getFilename() end
  if key == 'url' then return self:getUrl() end
  if key == 'path' then return storage..'/'..self.filename end
  if key == 'temp' then return self.path..'.tmp' end
  return getmetatable(self)[key]
end

-- Return true, if the this package is newer then `pack' package
-- Only versions are checked, names are ignored
function Package:isNewerThan(pack)
  return compare_ver(self.ver, pack.ver) > 0
end

-- Return true, if the this package is older then `pack' package
-- Only versions are checked, names are ignored
function Package:isOlderThan(pack)
  return compare_ver(self.ver, pack.ver) < 0
end

-- Operator for compare function (==, ~=)
-- Names have to be the same for everything but firmware
function Package:__eq(pack)
  if self.type == "cdi" and pack.type == "cdi" then
    return compare_ver(self.ver, pack.ver) == 0
  end
  return self.name == pack.name and compare_ver(self.ver, pack.ver) == 0
end

-- Operator for compare function (<, >,)
-- Names have to be the same for everything but firmware
function Package:__lt(pack)
  if self.type == "cdi" and pack.type == "cdi" then
    return compare_ver(self.ver, pack.ver) < 0
  end
  return self.name == pack.name and compare_ver(self.ver, pack.ver) < 0
end

-- Operator for compare function (<=, >=)
-- Names have to be the same for everything but firmware
function Package:__le(pack)
  if self.type == "cdi" and pack.type == "cdi" then
    return compare_ver(self.ver, pack.ver) <= 0
  end
  return self.name == pack.name and compare_ver(self.ver, pack.ver) <= 0
end

-- Full name of packages name (filename)
function Package:getFilename()
  return self.name .. "." .. self.type
end

-- Print package as string
function Package:__tostring()
  return self.name..'-'..self.ver
end

-- URL to download from
function Package:getUrl()
  if self._url == nil then return repo..'/'..self:getFilename() end
  if self._url:match('://') then return self._url end
  return repo..'/'..self._url
end

-- Download package from "repo" to "storage" (global variables)
function Package:download()
  print("DOWNLOAD call", self.name)
  if not self.downloadFinished then
    if self.conn == nil then
      if file_exists(self.path) then
        if not os.remove(self.path) then
          print('Can not remove old file:', self.path)
          self.downloadError = true
          return false
        end
      end
      self.conn = c8y.net:get(self.url)
      self.conn:setOutFile(self.temp)
      self.conn:send()
      print("Downloading", self.name, "from", self.url, "to", self.temp)
    end
    if not self.conn.isFinished then return false end
    if self.conn.ok then
      if os.rename(self.temp, self.path) then
        print("Downloading", self.name, "finished")
        self.downloadFinished = true
        self.conn:close()
        return true
      end
      print('Can not rename file '..self.temp..' to '..self.filename)
      self.conn:close()
      self.downloadError = true
      return false
    end
    self.downloadError = true
    self.conn:close()
    print('Downloading ' .. self.filename .. ' failed')
    return false
  else
    return true
  end
end
-----------------------------------------------------------------------------
-- Packages class (list of package)
class "Packages"

-- Package list constructor
function Packages:Packages()
end

-- Packege list constructor from installed packages (ipkg-cl) also adds the firmware package
function Packages.fromInstalled()
  local res = command('list_installed')
  --  local res = "curl - 7.37.0 -\nlibconfig - 1.4.9 -\nsmartrest-agent - 1.4.1 -\n"
  local out = Packages()
  res = res:split_simple('\n')
  for _, ln in pairs(res) do
    ln = ln:split_simple()
    local p = Package(ln[1], ln[3], nil, "ipk")
    out[p.name] = p
  end
  return out
end

-- Packages iterators (returns only packages from list)
-- You can use pattern else you get all packages, e.g.:
--
-- for p in packs:packages() do
--   print(p.name)
-- done
--
function Packages:packages(pattern)
  local k = next(self)
  return function()
    while k ~= nil do
      local p = self[k]
      k = next(self, k)
      if p.__type == 'Package' and (pattern == nil or p.name:find(pattern)) then
        return p
      end
    end
  end
end

-- Add packages to list
function Packages:add(p)
  self[p.name] = p
end

-- Create and add new package
function Packages:create(name, ver, url, type)
  local p = Package(name, ver, url, type)
  self:add(p)
  return p
end

-- True, if list contains the package
function Packages:contains(p)
  return self[p.name] ~= nil
end

-- Gets the package with the specified name or nil no package with the name can be found
function Packages:get(name)
  if self[name] then
    return self[name]
  end
  return self[name:split_simple(".")[1]]
end

-- Compare two packages lists. Returns three packages lists (toInstall, toRemove, toUpdate)
function Packages:upgrade_list(newPackages)
  local toInstall = Packages()
  local toUpdate = Packages()
  local toRemove = Packages()
  -- find all new and newer packages
  for p in newPackages:packages() do
    if not self:contains(p) then
      toInstall:add(p)
    elseif p ~= self[p.name] then
      toUpdate:add(p)
    end
  end
  -- find removed packages
  for p in self:packages() do
    if not newPackages:contains(p) then
      toRemove:add(p)
    end
  end
  --
  return toInstall, toRemove, toUpdate
end

-- check if two packages lists are equal. Checks all contained packages
function Packages:__eq(other)
  for key, package in pairs(self) do
    if package ~= other:get(package.name) then
      return false
    end
  end

  for key, package in pairs(other) do
    if package ~= self[package.name] then
      return false
    end
  end
  return true
end
-----------------------------------------------------------------------------
-- The package manager
--
-- Create new manager in start() function:
--   manager = Manager()
--
-- When you have new packages use:
--   if not manager.busy then
--     manager:update(packages)
--     manager:upgrade()
--   end
--
-- And in your run() function always call:
--   manager:run()
--
class "Manager"

-- New package manager
function Manager:Manager()
  self.list = Packages.fromInstalled()
  self.new = Packages()
  self.busy = false
  self.tries = 3
  self.error = nil
  self.commandManager = CommandManager()
end

-- Update new packages. Create list of packeaget to remove and to install
function Manager:update(newPackages)
  self:_start()
  self.new = newPackages
  self.toInstall, self.toRemove , self.toUpdate = self.list:upgrade_list(self.new)
  self:_stop()
end

-- Download missing packages worker
function Manager:_download()
  os.execute('mkdir -p "' .. storage .. '"')
  local function downloadPackages(packages)
    for _, p in pairs(packages) do
      if not p:download() then
        if p.downloadError then
          self.tries = self.tries - 1
          if self.tries <= 0 then
            self.error = 'Downloading ' .. p.name .. ' failed'
          end
        end
        return false
      end
    end
    return true
  end
  if not downloadPackages(self.toInstall) then
    return false
  end
  if not downloadPackages(self.toUpdate) then
    return false
  end
  return true
end

-- Upgrade packages worker
function Manager:_upgrade()
  if not self:_download() then return false end
  local command = ""
  local lastCommand = ""

  for p in self.toRemove:packages() do
    local cmd, lateCmd = self.commandManager:remove(p)
    command = command .. cmd
    lastCommand = lastCommand .. lateCmd
  end
  for p in self.toInstall:packages() do
    local cmd, lateCmd = self.commandManager:install(p)
    command = command .. cmd
    lastCommand = lastCommand .. lateCmd
  end
  for p in self.toUpdate:packages() do
    local cmd, lateCmd = self.commandManager:update(p)
    command = command .. cmd
    lastCommand = lastCommand .. lateCmd
  end
  command = command .. lastCommand
  if not command:match("reboot") then
    command = command .. "reboot"
  end
  print(command)
  os.execute(command)
  return true
end

-- Start a new work, cmd is function to call in run()
function Manager:_start(cmd)
  if self.cmd == nil then
    self.cmd = cmd or function() return false end
    self.busy = true
    self.error = nil
    self.tries = 3
  else
    error('Already working')
  end
end

-- Stop the work
function Manager:_stop(cmd)
  if self.cmd == nil then return end
  self.cmd = nil
  self.busy = false
  print('Manager work finished')
end

-- Download missing packages
function Manager:download()
  self:_start(self._download)
end

-- Upgrade packages
function Manager:upgrade()
  self:_start(self._upgrade)
end

-- Continue manager work
function Manager:run()
  if self.cmd == nil then return end
  if self:cmd() then
    self:_stop()
  end
  if self.error then
    print('Manager ERROR: ' .. self.error)
    if self.op then
    	self.op:setStatus(c8y.operations.Failed, self.error)
    	self.op = nil
    end
    self.cmd = nil
    self.busy = false
    self.error = nil
    for _, package in pairs(self.toInstall) do
    	os.remove(package.path)
    	os.remove(package.temp)
    end
    for _, package in pairs(self.toUpdate) do
      os.remove(package.path)
      os.remove(package.temp)
    end
    os.remove(swOpBackup)
  end
end

-----------------------------------------------------------------------------
-- The c8y_SoftwareList handler
class "SoftwareOperation"
--
function SoftwareOperation:operationExecuting(op)
  print("EXECUTING SW OP", op.id)
  print(c8y.dump(op.records))
  local packs = Packages()
  if file_exists(swOpBackup) then
    local file = io.open(swOpBackup)
    local operationId = file:read("*n")
    file:close()

    if operationId ~= op.id then
      op:setStatus(c8y.operations.Failed, "Software update already in progress")
    end
  else
    local file = io.open(swOpBackup, "w+")
    file:write(op.id .. ",sw\n")
    for i=1, op.responseCount do
      local record = op:getResponseRecord(i)
      local name, version, url = record[3],record[4],record[5]
      local packageName = name:match("^[a-zA-Z0-9+-]+")
      if not packageName then
      	file:close()
      	op:setStatus(c8y.operations.Failed, "invalid package name specified: " .. (name or '""'))
      	os.remove(swOpBackup)
      end
      local p = packs:create(packageName, version, url, "ipk")
      local recordString = packageName .. "," .. version .. "," .. (url or " ") .. "," .. p.path .. "\n"
      file:write(recordString)
    end
    file:flush()
    file:close()

    manager.op = op
    manager:update(packs)
    manager:upgrade()
		op:stopTimeout()
  end
  return true
end
-----------------------------------------------------------------------------
-- the c8y_Firmware handler
class "FirmwareOperation"

function FirmwareOperation:operationExecuting(op)
  print("EXECUTING FW OP", op.id)
  if file_exists(swOpBackup) then
    local file = io.open(swOpBackup)
    local operationId = file:read("*n")
    file:close()

    if operationId ~= op.id then
      op:setStatus(c8y.operations.Failed, "Firmware update already in progress")
    end
  else
    local record = op:getResponseRecord(1)
    local name, version, url = record[3],record[4],record[5]
    print(name, version, url)
    local fwVersion = io.popen("rdb_get sw.version"):read("*l")
    if compare_ver(fwVersion, version) ~= 0 then
      local file = io.open(swOpBackup, "w+")
      file:write(op.id .. ",fw\n")

      local packages = Packages:fromInstalled()
      local p = packages:create(name, version, url, "cdi")

      local recordString = name .. "," .. version .. "," .. (url or " ") .. "," .. p.path .. "\n"
      file:write(recordString)
      file:flush()
      file:close()

      manager:update(packages)
      manager:upgrade()
			op:stopTimeout()
    else
      op:setStatus(c8y.operations.Successful)
    end
  end
end

-----------------------------------------------------------------------------

function init()
  manager = Manager()
  -- register templates
  templates = {}
  -- register operation handler fow c8y_SoftwareList
  initOperations()

  fwVersion = io.popen("rdb_get sw.version"):read("*l")
  if file_exists(swOpBackup) then
    print("verifying software...")
    local lines = io.lines(swOpBackup)
    local splitFirstLine = (lines() or ""):split_simple(",")
    local id = tonumber(splitFirstLine[1])
    local operationType = splitFirstLine[2]
    if not id or not operationType then
      os.remove(swOpBackup)
      return
    end
    local packages = Packages()
    if operationType == "sw" then
      for line in lines do
        local values = line:split_simple(",")
        local name, version, url, actualFile = values[1], values[2], values[3], values[4]
        local p = packages:create(name, version, url , "ipk")
        p.actualFile = actualFile
        print(name, version)
      end
      if manager.list == packages then
        prepareSavedOperation(id, c8y.operations.Successful)
      else
        prepareSavedOperation(id, c8y.operations.Failed, "Package verification failed")
      end
    elseif operationType == "fw" then
      local values = lines():split_simple(",")
      local name, version, url, actualFile = values[1], values[2], values[3], values[4]
      local package = packages:create(name, version, url, "cdi")
      package.actualFile = actualFile
      if fwVersion and (compare_ver(fwVersion, version) == 0) then
        prepareSavedOperation(id, c8y.operations.Successful)
      else
        prepareSavedOperation(id, c8y.operations.Failed, "Firmware version verification failed")
      end
    end
    print("removing obsolete file")
    os.remove(swOpBackup)
    for _, package in pairs(packages) do
      print("removing:", package.actualFile)
      os.remove(package.actualFile)
      os.remove(package.actualFile .. ".tmp")
    end
  end
end


function postInit()
  sendSavedOperationStatus()

  sendInstalledSoftware()
  sendInstalledFirmware(fwVersion or "0.0")
end

function run()
  manager:run()
end

-- Subbscribe templates and our handle for c8y_SoftwareList operation
function initOperations()
  -- register tempaltes
  templates.updateSoftware = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,,"{""c8y_SoftwareList"":[%%]}"')
  templates.updateFirmware = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,,"{""c8y_Firmware"":{""name"":""firmware.cdi"",""version"":""%%"",""url"":"" "",""moduleVersion"":""%%""}}"')
  -- register operation handler for c8y_SoftwareList
  swOperation = SoftwareOperation()
  c8y.operations:addOperation('c8y_SoftwareList', '$.name,$.version,$.url', swOperation)
  fwOperation = FirmwareOperation()
  c8y.operations:addOperation('c8y_Firmware', '$.name,$.version,$.url', fwOperation)
end

function sendInstalledSoftware()
  local s = ""
  DUMP(manager.list)
  for _, package in pairs(manager.list) do
    if #s > 0 then s = s .. ',' end
    s = s .. '{"name":"'..package.name..'","version":"'..package.ver..'","url":""}'
  end
  local res = templates.updateSoftware:send({c8y.device, s, JsonEscape=false})
  print('Update software:', res.ok, res.code, res.httpCode)
end

function sendInstalledFirmware(version)
  print(version)

	local moduleVersion = io.popen("rdb_get wwan.0.firmware_version"):read("*l")
  local res = templates.updateFirmware:send({c8y.device, version, moduleVersion})
  print('Update firmware:', res.ok, res.code, res.httpCode)
end

-------------------------------------------------------------------------
class "CommandManager"
function CommandManager:CommandManager()
end

function CommandManager:install(package)
  if package.type == "cdi" then
    return "", "install_file " .. package.path .. " reboot ; "
  end
  return "install_file " .. package.path .. " ; ", ""
end

function CommandManager:update(package)
  if package.type == "ipk" then
    return "ipkg-cl install " .. package.path .. " -force-downgrade ; " , ""
  else
    return self:install(package)
  end
end

function CommandManager:remove(package)
  if package.type == "ipk" then
    return "ipkg-cl remove " .. package.name .. " ; ", ""
  else
    return "rm " .. package.path .. " ; ", ""
  end
end
